# This file manages which languages are supported for OCR operations, not the user interface.
import sys
from enum import auto
from typing import Callable

from natsort import humansorted

# If using Python 3.10 or older, use the 3rd party StrEnum.
if sys.version_info < (3, 11):
    from strenum import StrEnum
else:
    from enum import StrEnum


class LanguageCode(StrEnum):
    detect_box = auto()
    detect_page = auto()
    jpn = auto()
    eng = auto()
    kor = auto()
    kor_vert = auto()
    chi_sim = auto()
    chi_tra = auto()
    sqi = auto()
    ara = auto()
    aze = auto()
    aze_cyrl = auto()
    ben = auto()
    bul = auto()
    mya = auto()
    cat = auto()
    hrv = auto()
    ces = auto()
    dan = auto()
    nld = auto()
    epo = auto()
    est = auto()
    fin = auto()
    fra = auto()
    kat = auto()
    deu = auto()
    ell = auto()
    heb = auto()
    hin = auto()
    hun = auto()
    ind = auto()
    ita = auto()
    kaz = auto()
    lat = auto()
    lit = auto()
    ltz = auto()
    msa = auto()
    mon = auto()
    nep = auto()
    nor = auto()
    fas = auto()
    pol = auto()
    por = auto()
    ron = auto()
    rus = auto()
    srp = auto()
    srp_latn = auto()
    slk = auto()
    slv = auto()
    spa = auto()
    swe = auto()
    tgl = auto()
    tam = auto()
    tel = auto()
    tha = auto()
    tur = auto()
    ukr = auto()
    vie = auto()

    def __str__(self):
        return self.value


def to_language_code(lang: str) -> LanguageCode | None:
    """
    Convert a language string to a LanguageCode enum.

    :param lang: The language string.
    :return: The LanguageCode enum or None if not found.
    """
    if lang in LanguageCode.__members__:
        return LanguageCode[lang]
    else:
        return None


LANGUAGE_CODE_TO_NAME: dict[LanguageCode, str] = {
    LanguageCode.detect_box: "Detect per box",
    LanguageCode.detect_page: "Detect per page",
    LanguageCode.jpn: "Japanese",
    LanguageCode.eng: "English",
    LanguageCode.kor: "Korean",
    LanguageCode.kor_vert: "Korean (vertical)",
    LanguageCode.chi_sim: "Chinese - Simplified",
    LanguageCode.chi_tra: "Chinese - Traditional",
    LanguageCode.sqi: "Albanian",
    LanguageCode.ara: "Arabic",
    LanguageCode.aze: "Azerbaijani",
    LanguageCode.aze_cyrl: "Azerbaijani - Cyrilic",
    LanguageCode.ben: "Bengali",
    LanguageCode.bul: "Bulgarian",
    LanguageCode.mya: "Burmese",
    LanguageCode.cat: "Catalan; Valencian",
    LanguageCode.hrv: "Croatian",
    LanguageCode.ces: "Czech",
    LanguageCode.dan: "Danish",
    LanguageCode.nld: "Dutch; Flemish",
    LanguageCode.epo: "Esperanto",
    LanguageCode.est: "Estonian",
    LanguageCode.fin: "Finnish",
    LanguageCode.fra: "French",
    LanguageCode.kat: "Georgian",
    LanguageCode.deu: "German",
    LanguageCode.ell: "Greek",
    LanguageCode.heb: "Hebrew",
    LanguageCode.hin: "Hindi",
    LanguageCode.hun: "Hungarian",
    LanguageCode.ind: "Indonesian",
    LanguageCode.ita: "Italian",
    LanguageCode.kaz: "Kazakh",
    LanguageCode.lat: "Latin",
    LanguageCode.lit: "Lithuanian",
    LanguageCode.ltz: "Luxembourgish",
    LanguageCode.msa: "Malay",
    LanguageCode.mon: "Mongolian",
    LanguageCode.nep: "Nepali",
    LanguageCode.nor: "Norwegian",
    LanguageCode.fas: "Persian",
    LanguageCode.pol: "Polish",
    LanguageCode.por: "Portuguese",
    LanguageCode.ron: "Romanian; Moldavian",
    LanguageCode.rus: "Russian",
    LanguageCode.srp: "Serbian",
    LanguageCode.srp_latn: "Serbian - Latin",
    LanguageCode.slk: "Slovak",
    LanguageCode.slv: "Slovenian",
    LanguageCode.spa: "Spanish; Castilian",
    LanguageCode.swe: "Swedish",
    LanguageCode.tgl: "Tagalog",
    LanguageCode.tam: "Tamil",
    LanguageCode.tel: "Telugu",
    LanguageCode.tha: "Thai",
    LanguageCode.tur: "Turkish",
    LanguageCode.ukr: "Ukrainian",
    LanguageCode.vie: "Vietnamese",
}

LANGUAGES_RTL_BOX_ORDER: set[LanguageCode] = {
    LanguageCode.jpn,
    LanguageCode.chi_sim,
    LanguageCode.chi_tra,
    LanguageCode.ara,
    LanguageCode.fas,
    LanguageCode.heb,
}


def language_code_name_sorted(
    include_detect: bool = True,
    pin_important: bool = True,
    translate: Callable[[str], str] = lambda x: x,
) -> list[tuple[LanguageCode, str]]:
    """
    Sort the languages by their (translated) name.

    :param include_detect: Include the detection settings as "languages".
    :param pin_important: Pin important languages to the top.
    :param translate: [Optional] The translation function.
    """
    if include_detect:
        langs = list(LanguageCode)
    else:
        langs = list(LanguageCode)
        langs.remove(LanguageCode.detect_box)
        langs.remove(LanguageCode.detect_page)

    important_langs = []
    if pin_important:
        # Pull out the important languages.
        important_langs = [
            LanguageCode.jpn,
            LanguageCode.eng,
            LanguageCode.kor,
            LanguageCode.kor_vert,
            LanguageCode.chi_sim,
            LanguageCode.chi_tra,
        ]
        if include_detect:
            important_langs = [LanguageCode.detect_box, LanguageCode.detect_page] + important_langs
        # Remove the important languages from the list.
        langs = [lang for lang in langs if lang not in important_langs]

    # Apply the translation.
    langs = [(lang, translate(LANGUAGE_CODE_TO_NAME[lang])) for lang in langs]

    sorted_langs = humansorted(langs, key=lambda x: x[1])

    if pin_important:
        # Put the important languages at the top.
        important_langs = [
            (lang, translate(LANGUAGE_CODE_TO_NAME[lang])) for lang in important_langs
        ]
        sorted_langs = important_langs + sorted_langs

    return sorted_langs



import yaml
from pathlib import Path, PosixPath
from importlib import resources

import pcleaner.gui.gui_utils as gu
import pcleaner.data.theme_icons as theme_icons_module


def test_get_available_themes():
    """
    Test the get_available_themes function.
    """
    themes = gu.get_available_themes()

    assert themes
    assert themes == [("breeze", "Breeze Light"), ("breeze-dark", "Breeze Dark")]


def test_theme_icon_presence():
    """
    Test that all icons are present in the resources.
    """
    # Read the icon list from the yaml file located at DeepQt/icons/theme_list.yaml
    yaml_path = Path(__file__).parent.parent / "icons" / "theme_list.yaml"
    with resources.files(theme_icons_module) as theme_icons_path:
        theme_icons_root = Path(theme_icons_path)

    with yaml_path.open() as file:
        data = yaml.safe_load(file)

    theme_directories = data["Theme directories"]
    theme_icons = data["Files"]

    # Perform a depth-first search for each icon in a theme, which means
    # checking for leaf nodes, which are the icon file names.
    # When coming across a directory, the name of that directory is added to the path,
    # then scanned recursively.
    for theme_path in theme_directories:
        # Only use the last directory name of the theme path.
        theme_directory_name = PosixPath(theme_path).name
        for category, sizes in theme_icons.items():
            for size, icons in sizes.items():
                for icon in icons:
                    expected_relative_path = (
                        PosixPath(theme_directory_name) / category / str(size) / icon
                    )

                    for suffix in ["svg", "png"]:
                        full_path = (
                            theme_icons_root
                            / expected_relative_path.with_suffix(f".{suffix}").as_posix()
                        )
                        if full_path.exists():
                            break
                    else:
                        assert (
                            False
                        ), f"Icon {expected_relative_path} not found among the theme_icons data files."



from pathlib import Path
from csv import reader
from enum import Enum, auto
from attrs import frozen
from collections import defaultdict

import pcleaner.structures as st


class ParseErrorCode(Enum):
    OS_ERROR = auto()
    OTHER_ERROR = auto()
    NOT_6_COLUMNS = auto()
    NOT_AN_INT = auto()
    INT_TOO_BIG = auto()
    INVALID_FORMAT = auto()
    INVALID_CSV_HEADER = auto()
    NO_FILE_PATH = auto()
    INVALID_PATH = auto()


@frozen
class ParseError:
    """
    Store enough information to inform the user about parse errors.
    Use error codes to allow for easy i18n.
    """

    line: int
    error_code: ParseErrorCode
    context: str


def parse_plain_text(path: Path) -> tuple[list[st.OCRAnalytic], list[ParseError]]:
    """
    Try to load the file, assuming it's a plain text file.
    If it isn't, raise a FileNotPlain Error, along with relevant
    parse errors.

    Valid example:

    myfile.jpg:
    This is some text.
    This is some more text.

    anotherfile.jpg:
    This is some text.

    :param path: The path to the supposed csv file.
    :return: A list of analytics and a list of errors, one of which will be empty.
    """

    analytics_data: dict[Path, list[tuple[str, st.Box]]] = defaultdict(list)
    parse_errors: list[ParseError] = []

    with path.open("r", encoding="utf-8") as file:
        lines = file.readlines()

    current_path = None
    expecting_file_path = True

    for line_number, line in enumerate(lines, start=1):
        stripped_line = line.strip()

        # Simple 2-state machine to parse the plain text file.
        if expecting_file_path:
            # Ignore consecutive empty lines.
            if not stripped_line:
                continue

            if len(stripped_line) > 1 and stripped_line.endswith(":"):
                current_path = Path(stripped_line[:-1])
                expecting_file_path = False
                continue
            else:
                parse_errors.append(
                    ParseError(
                        line=line_number, error_code=ParseErrorCode.INVALID_PATH, context=line
                    )
                )
                continue

        else:
            if not stripped_line:
                expecting_file_path = True
                continue
            else:
                text = stripped_line
                box = st.Box(-1, -1, -1, -1)
                analytics_data[current_path].append((text, box))

    if parse_errors:
        return [], parse_errors

    # Pack the analytics data by file path.
    analytics_list = []
    for file_path, box_data in analytics_data.items():
        if not box_data:
            continue
        analytics_list.append(
            st.OCRAnalytic(
                len(box_data), [], [], tuple((file_path, text, box) for text, box in box_data)
            )
        )

    return analytics_list, []


def parse_csv(path: Path) -> tuple[list[st.OCRAnalytic], list[ParseError]]:
    """
    Try to load the file, assuming it's a csv file.
    If it isn't, raise a FileNotCSV Error, along with relevant
    parse errors.

    Valid example:

    filename,startx,starty,endx,endy,text
    img1.jpg,923,73,1011,336,sample text
    img1.jpg,423,73,711,336,"something else, perhaps"
    img2.jpg,534,275,592,414,or nothing at all

    :param path: The path to the supposed csv file.
    :return: A list of analytics and a list of errors, one of which will be empty.
    """
    analytics_data: dict[Path, list[tuple[str, st.Box]]] = defaultdict(list)
    parse_errors: list[ParseError] = []

    with path.open("r") as file:
        csv_reader = reader(file)
        header = next(csv_reader)

        # The header must not contain any digits.
        # If it does, then someone likely removed the header and now has box coordinates
        # in here, which would get ignored.
        if len(header) != 6 or any(h.isdigit() for h in header):
            parse_errors.append(
                ParseError(
                    line=1,
                    error_code=ParseErrorCode.INVALID_CSV_HEADER,
                    context=",".join(header),
                )
            )
            return [], parse_errors

        for line_number, row in enumerate(csv_reader, start=2):
            # Each row needs 6 columns.
            if len(row) != 6:
                parse_errors.append(
                    ParseError(
                        line=line_number,
                        error_code=ParseErrorCode.NOT_6_COLUMNS,
                        context=",".join(row),
                    )
                )
                continue

            filename, startx, starty, endx, endy, text = row

            try:
                startx = int(startx)
                starty = int(starty)
                endx = int(endx)
                endy = int(endy)
            except ValueError:
                parse_errors.append(
                    ParseError(
                        line=line_number,
                        error_code=ParseErrorCode.NOT_AN_INT,
                        context=",".join(row),
                    )
                )
                continue
            # Make sure none of the ints is too big for a signed 32-bit int,
            # otherwise Qt can't render the boxes.
            if any(abs(i) > 2**31 - 1 for i in (startx, starty, endx, endy)):
                parse_errors.append(
                    ParseError(
                        line=line_number,
                        error_code=ParseErrorCode.INT_TOO_BIG,
                        context=",".join(row),
                    )
                )
                continue

            # Ensure the x and y values are coordinates, such that
            # startx <= endx (analogous for y). Otherwise they don't
            # form a valid box.
            startx, endx = min(startx, endx), max(startx, endx)
            starty, endy = min(starty, endy), max(starty, endy)

            if not filename:
                parse_errors.append(
                    ParseError(
                        line=line_number,
                        error_code=ParseErrorCode.NO_FILE_PATH,
                        context=",".join(row),
                    )
                )
                continue

            try:
                file_path = Path(filename)
            except Exception:
                parse_errors.append(
                    ParseError(
                        line=line_number,
                        error_code=ParseErrorCode.INVALID_PATH,
                        context=",".join(row),
                    )
                )
                continue

            box = st.Box(startx, starty, endx, endy)

            analytics_data[file_path].append((text, box))

    if parse_errors:
        return [], parse_errors

    # Pack the analytics data by file path.
    analytics_list = []
    for file_path, box_data in analytics_data.items():
        analytics_list.append(
            st.OCRAnalytic(
                len(box_data), [], [], tuple((file_path, text, box) for text, box in box_data)
            )
        )

    return analytics_list, []


def parse_ocr_data(path: Path) -> tuple[list[st.OCRAnalytic], list[ParseError]]:
    """
    Parse the OCR data from the file at the given path.
    The file can be either a plain text file or a CSV file.

    :param path: The path to the OCR data file.
    :return: A list of analytics and a list of errors.
    """
    try:
        if path.suffix == ".csv":
            return parse_csv(path)
        elif path.suffix == ".txt":
            return parse_plain_text(path)
        else:
            return [], [ParseError(line=-1, error_code=ParseErrorCode.INVALID_FORMAT, context="")]
    except OSError as e:
        return [], [ParseError(line=-1, error_code=ParseErrorCode.OS_ERROR, context=str(e))]
    except Exception as e:
        return [], [ParseError(line=-1, error_code=ParseErrorCode.OTHER_ERROR, context=str(e))]



import sys
import winreg
import pyuac


def checkWindows():
    return "win32" in sys.platform


def generateKeys(programPath: str):
    createContextMenuTask(r"\\Directory\\Background\\shell\\", programPath)
    createContextMenuTask(r"\\Directory\\shell\\", programPath)

    createContextMenuSubCommand("Pclean.both", programPath + ' clean "%V" --notify', "Save both masks and cleaned page")
    createContextMenuSubCommand("Pclean.mask", programPath + ' clean -m "%V" --notify', "Save only masks")
    createContextMenuSubCommand("Pclean.clean", programPath + ' clean -c "%V" --notify', "Save only cleaned page")


def createContextMenuTask(path: str, programPath: str):
    currentKey = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, path, access=winreg.KEY_ALL_ACCESS)

    PcleanKey = winreg.CreateKey(currentKey, "Pcleaner")

    winreg.SetValueEx(PcleanKey, r"Icon", 0, winreg.REG_SZ, programPath)
    winreg.SetValueEx(PcleanKey, r"MUIVerb", 0, winreg.REG_SZ, r'Run Panel Cleaner against this folder')
    winreg.SetValueEx(PcleanKey, r"SubCommands", 0, winreg.REG_SZ, r'Pclean.both;|;Pclean.mask;Pclean.clean;')

    winreg.CloseKey(PcleanKey)
    winreg.CloseKey(currentKey)


def createContextMenuSubCommand(name: str, command: str, description: str):
    currentKey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                r"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\CommandStore\\shell",
                                access=winreg.KEY_ALL_ACCESS)

    cleanTaskKey = winreg.CreateKey(currentKey, name)

    winreg.SetValueEx(cleanTaskKey, r"MUIVerb", 0, winreg.REG_SZ, description)

    commandKey = winreg.CreateKey(cleanTaskKey, "command")

    winreg.SetValueEx(commandKey, "", 0, winreg.REG_SZ, command)

    winreg.CloseKey(commandKey)
    winreg.CloseKey(cleanTaskKey)
    winreg.CloseKey(currentKey)


def deleteKeys():
    deleteSubCommandKey("Pclean.both")
    deleteSubCommandKey("Pclean.clean")
    deleteSubCommandKey("Pclean.mask")

    deleteContextMenuKey()


def deleteContextMenuKey():
    try:
        currentKey = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "\\Directory\\Background\\shell\\Pcleaner",
                                    access=winreg.KEY_ALL_ACCESS)

        winreg.DeleteValue(currentKey, "MUIVerb")
        winreg.DeleteValue(currentKey, "Icon")
        winreg.DeleteValue(currentKey, "SubCommands")

        winreg.DeleteKey(currentKey, "")
        winreg.CloseKey(currentKey)

        currentKey = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "\\Directory\\shell\\Pcleaner",
                                    access=winreg.KEY_ALL_ACCESS)

        winreg.DeleteValue(currentKey, "MUIVerb")
        winreg.DeleteValue(currentKey, "Icon")
        winreg.DeleteValue(currentKey, "SubCommands")

        winreg.DeleteKey(currentKey, "")
        winreg.CloseKey(currentKey)
    except FileNotFoundError:
        return


def deleteSubCommandKey(keyName: str):
    try:
        currentKey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                    r"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\CommandStore\\shell\\" + keyName,
                                    access=winreg.KEY_ALL_ACCESS)

        commandKey = winreg.OpenKey(currentKey, "command")

        winreg.DeleteKey(commandKey, "")
        winreg.CloseKey(commandKey)

        winreg.DeleteValue(currentKey, "MUIVerb")
        winreg.DeleteKey(currentKey, "")
        winreg.CloseKey(currentKey)
    except FileNotFoundError:
        return


if __name__ == "__main__":
    if len(sys.argv) == 3:

        if sys.argv[1] == "create" or sys.argv[1] == "delete":
            if not pyuac.isUserAdmin():
                print("Re-launching as admin!")
                pyuac.runAsAdmin()
            else:
                if sys.argv[1] == "create":
                    generateKeys(sys.argv[2])
                elif sys.argv[1] == "delete":
                    deleteKeys()
        else:
            print("Invalid command, please enter create or delete followed by the path of the Panel Cleaner Exe")
            sys.exit(1)
    else:
        print("Invalid command, please enter create or delete followed by the path of the Panel Cleaner Exe")
        sys.exit(1)

    print("Finished")



import threading
import time

import torch

import pcleaner.helpers as hp


def memory_watcher():
    # Monitor system memory usage to warn of impending OOM errors.

    while True:
        # Print warnings based on RAM and swap status.
        if hp.sys_swap_memory_total() == 0:  # No swap available.
            if (mem := hp.sys_virtual_memory_used_percent()) >= 80:
                print_warning_banner(f"Warning: RAM usage has reached {round(mem)}%")
        else:
            if (mem := hp.sys_virtual_memory_used_percent()) >= 90:
                print_warning_banner(f"Warning: RAM usage has reached {round(mem)}%.")

        # Monitor VRAM if torch is available and GPU is detected.
        if torch.cuda.is_available() and torch.cuda.device_count() > 0:
            vram_total = torch.cuda.get_device_properties(0).total_memory
            vram_used = torch.cuda.memory_allocated(0)
            vram_percent = (vram_used / vram_total) * 100

            if vram_percent >= 80:
                print_warning_banner(f"Warning: VRAM usage has reached {round(vram_percent)}%")

        time.sleep(5)


def start_memory_watcher():
    watcher_thread = threading.Thread(target=memory_watcher, daemon=True)
    watcher_thread.start()


def print_warning_banner(message: str) -> None:
    trim = "=" * len(message)
    print("\n\033[91m" + trim + "\n" + message + "\n" + trim + "\033[0m\n")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'OCRLanguageSupport.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QDialog, QFrame,
    QHBoxLayout, QHeaderView, QPushButton, QSizePolicy,
    QSpacerItem, QTableWidgetItem, QVBoxLayout, QWidget)

from pcleaner.gui.CustomQ.CTableWidget import CTableWidget

class Ui_OCRLanguageSupport(object):
    def setupUi(self, OCRLanguageSupport):
        if not OCRLanguageSupport.objectName():
            OCRLanguageSupport.setObjectName(u"OCRLanguageSupport")
        OCRLanguageSupport.resize(500, 720)
        self.verticalLayout = QVBoxLayout(OCRLanguageSupport)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 0, 1, -1)
        self.lang_table = CTableWidget(OCRLanguageSupport)
        if (self.lang_table.columnCount() < 4):
            self.lang_table.setColumnCount(4)
        __qtablewidgetitem = QTableWidgetItem()
        self.lang_table.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.lang_table.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.lang_table.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.lang_table.setHorizontalHeaderItem(3, __qtablewidgetitem3)
        self.lang_table.setObjectName(u"lang_table")
        self.lang_table.setFocusPolicy(Qt.NoFocus)
        self.lang_table.setFrameShape(QFrame.NoFrame)
        self.lang_table.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.lang_table.setProperty("showDropIndicator", False)
        self.lang_table.setDragDropOverwriteMode(False)
        self.lang_table.setDragDropMode(QAbstractItemView.NoDragDrop)
        self.lang_table.setAlternatingRowColors(True)
        self.lang_table.setSelectionMode(QAbstractItemView.MultiSelection)
        self.lang_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.lang_table.setSortingEnabled(False)
        self.lang_table.setCornerButtonEnabled(False)
        self.lang_table.horizontalHeader().setVisible(True)
        self.lang_table.horizontalHeader().setMinimumSectionSize(80)
        self.lang_table.horizontalHeader().setHighlightSections(False)
        self.lang_table.verticalHeader().setVisible(False)
        self.lang_table.verticalHeader().setHighlightSections(False)

        self.verticalLayout.addWidget(self.lang_table)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(6, 6, -1, -1)
        self.pushButton_install_tesseract = QPushButton(OCRLanguageSupport)
        self.pushButton_install_tesseract.setObjectName(u"pushButton_install_tesseract")
        icon = QIcon()
        iconThemeName = u"internet-services"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_install_tesseract.setIcon(icon)

        self.horizontalLayout.addWidget(self.pushButton_install_tesseract)

        self.pushButton_install_tesseract_language_packs = QPushButton(OCRLanguageSupport)
        self.pushButton_install_tesseract_language_packs.setObjectName(u"pushButton_install_tesseract_language_packs")
        self.pushButton_install_tesseract_language_packs.setIcon(icon)

        self.horizontalLayout.addWidget(self.pushButton_install_tesseract_language_packs)

        self.horizontalSpacer = QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(OCRLanguageSupport)

        QMetaObject.connectSlotsByName(OCRLanguageSupport)
    # setupUi

    def retranslateUi(self, OCRLanguageSupport):
        OCRLanguageSupport.setWindowTitle(QCoreApplication.translate("OCRLanguageSupport", u"OCR Language Support", None))
        ___qtablewidgetitem = self.lang_table.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("OCRLanguageSupport", u"Code", None));
        ___qtablewidgetitem1 = self.lang_table.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("OCRLanguageSupport", u"Language", u"Can also call this statistics."));
        ___qtablewidgetitem2 = self.lang_table.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("OCRLanguageSupport", u"MangaOCR", None));
        ___qtablewidgetitem3 = self.lang_table.horizontalHeaderItem(3)
        ___qtablewidgetitem3.setText(QCoreApplication.translate("OCRLanguageSupport", u"Tesseract", None));
        self.pushButton_install_tesseract.setText(QCoreApplication.translate("OCRLanguageSupport", u"How to install Tesseract", None))
        self.pushButton_install_tesseract_language_packs.setText(QCoreApplication.translate("OCRLanguageSupport", u"How to install Tesseract language packs", None))
    # retranslateUi




from pathlib import Path
from importlib import resources

import PySide6.QtCore as Qc
from PySide6.QtCore import Qt
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Signal, Slot
from loguru import logger

import pcleaner.data
import pcleaner.gui.structures as gst

ZOOM_TICK_FACTOR = 1.25


class ImageViewer(Qw.QGraphicsView):
    mouseMoved = Qc.Signal(int, int)
    image_dimensions: tuple[int, int] | None
    image_item: Qw.QGraphicsPixmapItem | None
    # Share the pixmap within the image item to optimize multiple viewers
    # of the same image (as with the diff viewers).
    shared_pixmap: gst.Shared[Qg.QPixmap | None] | None

    loaded_image_path: Path | None

    # We want to emit when the zoom factor changes,
    # and also receive them.
    zoom_factor_changed = Signal(float)
    # Slot: set_zoom_factor = Slot(float)

    def __init__(self, parent=None) -> None:
        super(ImageViewer, self).__init__(parent)
        self.setRenderHint(Qg.QPainter.Antialiasing)

        # Scene for QGraphicsView
        self.scene = Qw.QGraphicsScene(self)
        self.setScene(self.scene)

        self.image_item = None
        self.image_dimensions = None
        self.loaded_image_path = None
        self.shared_pixmap = None
        # Disable the allocation limit for high resolution images.
        Qg.QImageReader.setAllocationLimit(0)

        self.zoom_factor = 1.0
        self.setAlignment(Qt.AlignCenter)

        # I'm confused as to what this is supposed to do.
        # Docs say this will prevent it from receiving mouse move events when no button
        # is pressed, but it somehow still does. Is it overridden somewhere??
        self.setMouseTracking(False)

        # Make zooming focus on mouse.
        self.setTransformationAnchor(Qw.QGraphicsView.ViewportAnchor.AnchorUnderMouse)

    def share_pixmap_container(self, container: gst.Shared) -> None:
        """
        Accept a shared container to deduplicate pixmap loads in memory.

        :param container: A sharable container.
        """
        self.shared_pixmap = container

    @Slot(float)
    def set_zoom_factor(self, factor: float) -> None:
        self.zoom_factor = factor
        self.zoom(1, suppress_signals=True)

    def pixmap_invalid(self) -> bool:
        return self.image_item is None or (
            self.image_item is not None and self.image_item.pixmap().isNull()
        )

    def load_cache_aware(self, image_path: Path) -> Qg.QPixmap:
        # If a container is provided, use that, or fill it if it's empty.
        if (
            self.shared_pixmap is None
            or self.shared_pixmap.is_none()
            or self.shared_pixmap.get().isNull()
        ):
            image = load_image_with_orientation(image_path)
            pixmap = Qg.QPixmap.fromImage(image)
            if self.shared_pixmap is not None:
                self.shared_pixmap.set(pixmap)
        else:
            pixmap = self.shared_pixmap.get()
        return pixmap

    def set_image(self, image_path: Path = None) -> None:
        self.loaded_image_path = image_path

        if image_path:
            self.scene.clear()
            self.image_item = Qw.QGraphicsPixmapItem()
            self.scene.addItem(self.image_item)
            # Move the image item by -0.5, -0.5 to make it align with the scene's
            self.image_item.setOffset(-0.5, -0.5)

            pixmap = self.load_cache_aware(image_path)
            self.image_item.setPixmap(pixmap)
            self.reset_scene_rect()
            dim = self.image_item.pixmap().size()
            self.image_dimensions = (dim.width(), dim.height())
            self.update_smoothing()
        else:
            # Display "nothing" message when no image is set
            self.scene.clear()
            self.image_dimensions = None
            self.image_item = None

    def reset_scene_rect(self) -> None:
        self.setSceneRect(self.image_item.pixmap().rect())

    def update_smoothing(self) -> None:
        # Do no smoothing if zooming is, so when a pixel covers more than 1 pixel.
        if self.zoom_factor > 1:
            self.setRenderHint(Qg.QPainter.SmoothPixmapTransform, False)
            self.set_transformation_mode(Qt.FastTransformation)
        else:
            self.setRenderHint(Qg.QPainter.SmoothPixmapTransform, True)
            self.set_transformation_mode(Qt.SmoothTransformation)

    def wheelEvent(self, event: Qg.QWheelEvent) -> None:
        if Qt.ControlModifier & event.modifiers():
            # Zoom in/out with Ctrl + mouse wheel
            if event.angleDelta().y() > 0:
                self.zoom_in(wheel=True)
            else:
                self.zoom_out(wheel=True)
        elif Qt.ShiftModifier & event.modifiers():
            # Horizontal scrolling with Shift + mouse wheel
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() - event.angleDelta().y()
            )
        else:
            # Default behavior for panning
            super().wheelEvent(event)

    # For the mouse wheel events, we want to zoom slower.
    # Since zooming applies a factor, we can use a root of the factor to
    # achieve half of a zoom step.
    def zoom_in(self, wheel=False) -> None:
        if wheel:
            self.zoom(ZOOM_TICK_FACTOR**0.5)
        else:
            self.zoom(ZOOM_TICK_FACTOR)

    def zoom_out(self, wheel=False) -> None:
        if wheel:
            self.zoom(1 / (ZOOM_TICK_FACTOR**0.5))
        else:
            self.zoom(1 / ZOOM_TICK_FACTOR)

    def zoom_reset(self) -> None:
        self.zoom_factor = 1.0
        self.zoom(1)

    def zoom_fit(self) -> None:
        if self.image_dimensions is None:
            return
        width, height = self.image_dimensions
        view_width, view_height = self.viewport().width(), self.viewport().height()
        width_factor = view_width / width
        height_factor = view_height / height
        self.zoom_factor = min(width_factor, height_factor)
        # logger.debug(
        #     f"Zoom to fit with factor: {self.zoom_factor:.2%} for view size {view_width}x{view_height} and image size {width}x{height}"
        # )
        self.zoom(1)

    def image_position(self, pos) -> Qc.QPoint:
        return self.mapToScene(pos).toPoint()

    def mouseMoveEvent(self, event) -> None:
        if self.pixmap_invalid():
            return
        # Call base class implementation for standard behavior
        super().mouseMoveEvent(event)
        # Emit the mouse position always
        image_pos = self.image_position(event.pos())
        self.mouseMoved.emit(image_pos.x(), image_pos.y())
        self.viewport().update()  # Request redraw for the pixel highlight

    def drawForeground(self, painter, rect) -> None:
        if not self.pixmap_invalid() and self.zoom_factor > 5:
            view_pos = self.mapFromGlobal(Qg.QCursor.pos())
            image_pos = self.mapToScene(view_pos).toPoint()

            # Check if the cursor is inside the pixmap area
            if self.image_item.pixmap().rect().contains(image_pos):
                # Decide the color for the square
                pixel_color = self.image_item.pixmap().toImage().pixelColor(image_pos)
                avg = (pixel_color.red() + pixel_color.green() + pixel_color.blue()) / 3
                square_color = Qt.white if avg < 128 else Qt.black

                painter.setPen(Qg.QPen(square_color, 1 / self.zoom_factor))
                rect = Qc.QRectF(image_pos.x() - 0.5, image_pos.y() - 0.5, 1, 1)
                scaled_rect = rect.adjusted(
                    -0.5 / self.zoom_factor,
                    -0.5 / self.zoom_factor,
                    0.5 / self.zoom_factor,
                    0.5 / self.zoom_factor,
                )
                painter.drawRect(scaled_rect)

    def drawBackground(self, painter, rect) -> None:
        super().drawBackground(painter, rect)  # Call the base class implementation
        if not self.pixmap_invalid():
            image_rect = self.image_item.boundingRect()
            bg_color = self.palette().highlight().color()
            # If the color is really dark, use a lighter color for the background and vice versa.
            if bg_color.lightness() < 40:
                bg_color = bg_color.lighter(110)
            else:
                bg_color = bg_color.darker(110)
            painter.setBrush(bg_color)
            painter.drawRect(image_rect)

    def zoom(self, factor: float, *, suppress_signals: bool = False) -> None:
        """
        Zoom the image by the given factor.
        The image may not exceed a scale of 100x or have both width and height be smaller
        than half of the viewport's width and height.

        In the even that this was called as a callback, we can suppress the signal.

        :param factor: The factor to multiply the current zoom factor by.
        :param suppress_signals: Whether to suppress the zoom factor changed signal.
        """
        proposed_zoom_factor = self.zoom_factor * factor
        proposed_zoom_factor = min(proposed_zoom_factor, 100)

        current_width, current_height = (
            self.image_item.pixmap().width(),
            self.image_item.pixmap().height(),
        )
        proposed_width, proposed_height = (
            current_width * proposed_zoom_factor,
            current_height * proposed_zoom_factor,
        )
        view_width, view_height = self.viewport().width(), self.viewport().height()

        # Don't zoom out further if it's getting too smol.
        if proposed_width < view_width / 2 and proposed_height < view_height / 2 and factor < 1:
            return

        self.zoom_factor = proposed_zoom_factor

        self.update_smoothing()
        self.setTransform(Qg.QTransform().scale(self.zoom_factor, self.zoom_factor))

        if not suppress_signals:
            self.zoom_factor_changed.emit(self.zoom_factor)

    def reset_zoom(self) -> None:
        self.zoom_factor = 1
        self.setTransform(Qg.QTransform().scale(self.zoom_factor, self.zoom_factor))
        self.zoom_factor_changed.emit(self.zoom_factor)

    def set_transformation_mode(self, mode: Qt.TransformationMode) -> None:
        """
        This function is broken out to allow subclasses with 2 images to
        apply the transformation mode to both images.

        :param mode: The transformation mode to apply.
        """

        self.image_item.setTransformationMode(mode)


class BubbleImageViewer(ImageViewer):
    """
    An ImageViewer that can draw (rectangular) bubbles on the image.
    """

    # These lists must all have the same length.
    bubbles: list[Qc.QRect]
    bubble_colors: list[Qg.QColor]
    bubble_labels: list[str]
    bubble_stroke: list[Qt.PenStyle]  # DashLine or SolidLine etc.

    allow_drawing_bubble: bool
    new_bubble_color: Qg.QColor

    _new_bubble_start: Qc.QPoint | None
    _new_bubble_end: Qc.QPoint | None
    new_bubble: Signal = Signal(Qc.QRect)
    bubble_clicked: Signal = Signal(int)  # The index of the bubble clicked.

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.bubbles = []
        self.bubble_colors = []
        self.bubble_labels = []
        self.bubble_stroke = []
        self.allow_drawing_bubble = False
        # Default new bubble color to palette highlight color.
        self.new_bubble_color = self.palette().highlight().color()

        self._new_bubble_start = None
        self._new_bubble_end = None

        # Load included font.
        with resources.files(pcleaner.data) as data_path:
            font_path = data_path / "LiberationSans-Regular.ttf"
        font_id = Qg.QFontDatabase.addApplicationFont(str(font_path))
        if font_id == -1:
            logger.error("Failed to load font.")
            self.font_name = None
        else:
            self.font_name = Qg.QFontDatabase.applicationFontFamilies(font_id)[0]

    def set_allow_drawing_bubble(self, allow: bool) -> None:
        self.allow_drawing_bubble = allow

    def set_new_bubble_color(self, color: Qg.QColor) -> None:
        self.new_bubble_color = color

    def clear_bubbles(self) -> None:
        self.bubbles.clear()
        self.bubble_colors.clear()
        self.bubble_labels.clear()
        self.bubble_stroke.clear()
        self.viewport().update()

    def set_bubbles(
        self,
        rects: list[Qc.QRect],
        colors: list[Qg.QColor],
        labels: list[str],
        strokes: list[Qt.PenStyle],
    ) -> None:
        self.bubbles = rects
        self.bubble_colors = colors
        self.bubble_labels = labels
        self.bubble_stroke = strokes
        self.viewport().update()

    def point_in_image(self, point: Qc.QPoint) -> bool:
        if self.image_item is None:
            return False
        return self.image_item.pixmap().rect().contains(point)

    def point_in_a_bubble(self, point: Qc.QPoint) -> int:
        """
        Figure out if any of the bubbles contain this point.

        :param point: The point to check.
        :return: The index of the bubble that contains the point, or -1 if none do.
        """
        for i, bubble in enumerate(self.bubbles):
            if bubble.contains(point):
                return i
        return -1

    def mousePressEvent(self, event: Qg.QMouseEvent) -> None:
        # Begin drawing a new bubble, if allowed.
        if self.allow_drawing_bubble:
            # Discard boxes that are started outside the image.
            if self.point_in_image(self.image_position(event.pos())):
                self._new_bubble_start = self.image_position(event.pos())
                self.setCursor(Qt.CrossCursor)
        else:
            # Check if a bubble was clicked.
            bubble_index = self.point_in_a_bubble(self.image_position(event.pos()))
            self.bubble_clicked.emit(bubble_index)
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: Qg.QMouseEvent) -> None:
        if self.allow_drawing_bubble and self._new_bubble_start is not None:
            self._new_bubble_end = self.image_position(event.pos())
            self.viewport().update()
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: Qg.QMouseEvent) -> None:
        # Finish drawing the bubble and emit the new bubble signal.
        if self.allow_drawing_bubble and self._new_bubble_start is not None:
            self._new_bubble_end = self.image_position(event.pos())
            if self.point_in_image(self._new_bubble_end):
                # Construct a positively sized QRect.
                # Qc.QRect(self._new_bubble_start, self._new_bubble_end)
                rect = Qc.QRect(
                    min(self._new_bubble_start.x(), self._new_bubble_end.x()),
                    min(self._new_bubble_start.y(), self._new_bubble_end.y()),
                    abs(self._new_bubble_end.x() - self._new_bubble_start.x()),
                    abs(self._new_bubble_end.y() - self._new_bubble_start.y()),
                )
                self.new_bubble.emit(rect)
                logger.debug(
                    f"New bubble drawn: {Qc.QRect(self._new_bubble_start, self._new_bubble_end)}"
                )
            self._new_bubble_start = None
            self._new_bubble_end = None
            self.viewport().update()
            self.setCursor(Qt.ArrowCursor)
        super().mouseReleaseEvent(event)

    def drawForeground(self, painter, rect) -> None:
        if self.image_item is None:
            return
        # Don't call super's function, we don't want the pixel highlight.

        # Use the included font for drawing bubble labels.
        font_size = int(self.image_dimensions[0] / 55) + 5  # Adjusted for scuffed Qt drawing.
        if self.font_name is not None:
            paint_font = Qg.QFont(self.font_name, font_size)
        else:
            paint_font = Qg.QFont("Arial", font_size)
        painter.setFont(paint_font)

        for bubble, color, label, stroke in zip(
            self.bubbles, self.bubble_colors, self.bubble_labels, self.bubble_stroke
        ):
            # Skip bubbles that have -1 for all coordinates. These are created when importing
            # plain text OCR data, which doesn't have box data.
            if (
                bubble.x() == -1
                and bubble.y() == -1
                and bubble.width() == 0
                and bubble.height() == 0
            ):
                continue
            # Draw the bubble with a solid outline and a fill at 48/255 opacity.
            painter.setPen(Qg.QPen(color, 1, stroke, j=Qt.PenJoinStyle.MiterJoin))
            painter.setBrush(Qg.QColor(color.red(), color.green(), color.blue(), 48))
            painter.drawRect(bubble)

            # Draw the label with a white outline.
            path = Qg.QPainterPath()
            path.addText(
                bubble.topLeft() - Qc.QPoint(int(-font_size * 0.1), int(font_size * 0.2)),
                paint_font,
                label,
            )

            outline_pen = Qg.QPen(Qt.white, 5, j=Qt.PenJoinStyle.RoundJoin)
            painter.setPen(outline_pen)
            painter.drawPath(path)

            text_pen = Qg.QPen(color)
            painter.setPen(text_pen)
            painter.fillPath(path, text_pen.color())

        if self._new_bubble_start is not None:
            # Use double the opacity for the new bubble to make it more visible.
            painter.setPen(Qg.QPen(self.new_bubble_color, 1, Qt.DashLine))
            painter.setBrush(
                Qg.QColor(
                    self.new_bubble_color.red(),
                    self.new_bubble_color.green(),
                    self.new_bubble_color.blue(),
                    48 * 2,
                )
            )
            painter.drawRect(Qc.QRect(self._new_bubble_start, self._new_bubble_end))


def load_image_with_orientation(image_path: Path | str) -> Qg.QImage:
    """
    This exists to accommodate images with embedded rotation tags.

    :param image_path: The image path.
    :return: A QImage object.
    """
    reader = Qg.QImageReader(str(image_path))
    image = reader.read()

    if not image.isNull():
        transformation = reader.transformation()
        # Create the appropriate QTransform
        if transformation == Qg.QImageIOHandler.TransformationMirror:
            transform = Qg.QTransform().scale(-1, 1)
        elif transformation == Qg.QImageIOHandler.TransformationFlip:
            transform = Qg.QTransform().scale(1, -1)
        elif transformation == Qg.QImageIOHandler.TransformationRotate180:
            transform = Qg.QTransform().rotate(180)
        elif transformation == Qg.QImageIOHandler.TransformationRotate90:
            transform = Qg.QTransform().rotate(90)
        elif transformation == Qg.QImageIOHandler.TransformationRotate270:
            transform = Qg.QTransform().rotate(270)
        else:
            transform = Qg.QTransform()

        # Apply the transformation if needed
        if not transform.isIdentity():
            image = image.transformed(transform, Qt.SmoothTransformation)

    return image



import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger

import pcleaner.gui.log_parser as lp
import pcleaner.gui.state_saver as ss
from pcleaner.gui.ui_generated_files.ui_ErrorDialog import Ui_ErrorDialog


class ErrorDialog(Qw.QDialog, Ui_ErrorDialog):
    """
    Show logs of individual sessions and allow the user to report an issue.
    """

    session_log: lp.LogSession

    state_saver: ss.StateSaver  # The state saver for the window.

    def __init__(
        self,
        parent=None,
        title: str = "Error",
        message: str = "An error occurred.",
    ) -> None:
        """
        Init the widget.

        :param parent: The parent widget.
        """
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.label_error_icon.setPixmap(Qg.QIcon.fromTheme("dialog-error").pixmap(64, 64))

        self.setWindowTitle(title)
        self.label_message.setText(message)

        self.pushButton_close.clicked.connect(self.close)
        self.pushButton_kill.clicked.connect(self.kill)
        self.pushButton_open_issues.clicked.connect(self.open_issues)
        self.pushButton_clipboard.clicked.connect(self.copy_to_clipboard)

        self.label_name_hidden.setText(
            self.tr('Note: Name "{name}" was hidden').format(name=lp.get_username())
        )
        self.load_session_log()

        self.state_saver = ss.StateSaver("error_dialog")
        self.init_state_saver()
        self.state_saver.restore()

    def closeEvent(self, event: Qg.QCloseEvent) -> None:
        """
        Called when the window is closed.
        """
        self.state_saver.save()
        event.accept()

    def init_state_saver(self) -> None:
        """
        Load the state from the state saver.
        """
        self.state_saver.register(
            self,
        )

    @staticmethod
    def open_issues() -> None:
        """
        Open the issues page in the browser.
        """
        logger.debug("Opening github issues page.")
        Qg.QDesktopServices.openUrl(Qc.QUrl("https://github.com/VoxelCubes/PanelCleaner/issues"))

    def load_session_log(self) -> None:
        """
        Load the logs.
        """
        log_text = lp.load_log_file()
        if log_text is None:
            return

        sessions = lp.parse_log_file(log_text, max_sessions=1)
        if len(sessions) != 1:
            self.session_log = lp.LogSession(self.tr("Failed to load log."))
        else:
            self.session_log = lp.parse_log_file(log_text, max_sessions=1)[0]

        self.log_viewer.show_log(self.session_log.text)
        self.log_viewer.scroll_to_bottom()

    def copy_to_clipboard(self) -> None:
        """
        Copy the issue report to the clipboard.
        """
        logger.debug("Copying issue report to clipboard.")
        Qg.QGuiApplication.clipboard().setText(self.session_log.text)

    def kill(self) -> None:
        """
        Kill the application.
        This may be necessary when repeated errors occur and the application is unresponsive.
        """
        logger.critical("User killed the application.")
        Qw.QApplication.instance().quit()  # Embrace death.
        self.close()
        raise SystemExit(255)



import PySide6.QtWidgets as Qw
from PySide6.QtCore import Qt
import PySide6.QtGui as Qg


# noinspection PyPep8Naming
class CTextEdit(Qw.QTextEdit):
    """
    A QTextEdit with a few extra features.
    - A context menu option to clear the text.
    """

    def __init__(self, parent=None) -> None:
        Qw.QTextEdit.__init__(self, parent)

        # Create a custom action for clearing the text
        self.clear_action = Qg.QAction(self.tr("Clear"), self)
        self.clear_action.setShortcut(Qg.QKeySequence("Ctrl+L"))
        self.clear_action.setIcon(Qg.QIcon.fromTheme("edit-clear-history"))
        self.clear_action.triggered.connect(self.clear)

        # Create a shortcut for the clear action
        self.clear_shortcut = Qg.QShortcut(Qg.QKeySequence("Ctrl+L"), self)
        self.clear_shortcut.setContext(Qt.WidgetWithChildrenShortcut)
        self.clear_shortcut.activated.connect(self.clear)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()

        menu.addAction(self.clear_action)

        # If the text is empty, disable the clear action.
        self.clear_action.setEnabled(bool(self.toPlainText()))

        menu.exec_(event.globalPos())



import io
import re
import sys
from types import TracebackType
from typing import Literal
from io import StringIO
from pathlib import Path
from importlib import resources

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger

from pcleaner.gui.error_dialog_driver import ErrorDialog
import pcleaner.gui.worker_thread as wt
import pcleaner.ocr.parsers as op
import pcleaner.gui.image_file as imf
import pcleaner.structures as st
from pcleaner.helpers import tr
from pcleaner.data import custom_icons
from pcleaner.data import color_themes


# For all show functions, pad the dialog message, so that the dialog is not too narrow for the window title.
MIN_MSG_LENGTH = 50


class SelectableMessageBox(Qw.QMessageBox):
    """
    Subclass the QMessageBox to make the text selectable.
    """

    def __init__(self, *args, **kwargs):
        super(SelectableMessageBox, self).__init__(*args, **kwargs)
        self.setTextInteractionFlags(Qc.Qt.TextSelectableByMouse)

        labels = self.findChildren(Qw.QLabel)
        for label in labels:
            label.setTextInteractionFlags(
                Qc.Qt.TextSelectableByMouse
                | Qc.Qt.LinksAccessibleByMouse
                | Qc.Qt.LinksAccessibleByKeyboard
            )


def show_exception(
    parent,
    title: str,
    msg: str,
    error_bundle: None | wt.WorkerError | tuple[type, BaseException, TracebackType] = None,
    collect_exception: bool = True,
) -> None:
    """
    Show an exception in a dialog along with logs.
    This automatically gathers the exception information from the current context
    or a given worker error object.

    You can also skip collecting the exception if you have already logged this separately and
    merely wish to open the Issue Reporter dialog for the user.

    :param parent: The parent widget.
    :param title: The title of the dialog.
    :param msg: The message to show.
    :param error_bundle: [Optional] A worker error or a tuple of exception information.
    :param collect_exception: [Optional] Whether to add exception information to the log.
    """

    if collect_exception:
        exception_type: type[BaseException]
        exception_value: BaseException
        exception_traceback: TracebackType

        if error_bundle is not None:
            if isinstance(error_bundle, tuple):
                exception_type, exception_value, exception_traceback = error_bundle
            elif isinstance(error_bundle, wt.WorkerError):
                exception_type = error_bundle.exception_type
                exception_value = error_bundle.value
                exception_traceback = error_bundle.traceback
            else:
                logger.error(f"Invalid error bundle: {error_bundle}")
                exception_type, exception_value, exception_traceback = sys.exc_info()
        else:
            exception_type, exception_value, exception_traceback = sys.exc_info()

        # Ignore the exception if it's a KeyboardInterrupt.
        if exception_type is KeyboardInterrupt:
            logger.warning("User interrupted the process.")
            return

        logger.opt(
            depth=1, exception=(exception_type, exception_value, exception_traceback)
        ).critical(msg)

    box = ErrorDialog(parent, title, msg)
    box.exec()


def show_critical(parent, title: str, msg: str, **kwargs) -> int:
    msg = msg.ljust(MIN_MSG_LENGTH)
    box = SelectableMessageBox(
        Qw.QMessageBox.Critical,
        title,
        msg,
        Qw.QMessageBox.Yes | Qw.QMessageBox.Abort,
        parent,
        **kwargs,
    )
    return box.exec()


def show_warning(parent, title: str, msg: str) -> None:
    msg = msg.ljust(MIN_MSG_LENGTH)
    box = SelectableMessageBox(Qw.QMessageBox.Warning, title, msg, Qw.QMessageBox.Ok, parent)
    box.exec()


def show_info(parent, title: str, msg: str) -> None:
    msg = msg.ljust(MIN_MSG_LENGTH)
    box = SelectableMessageBox(Qw.QMessageBox.Information, title, msg, Qw.QMessageBox.Ok, parent)
    box.exec()


def show_question(
    parent, title: str, msg: str, buttons=Qw.QMessageBox.Yes | Qw.QMessageBox.Cancel
) -> int:
    msg = msg.ljust(MIN_MSG_LENGTH)
    dlg = Qw.QMessageBox(parent)
    dlg.setWindowTitle(title)
    dlg.setText(msg)
    dlg.setStandardButtons(buttons)
    dlg.setIcon(Qw.QMessageBox.Question)
    return dlg.exec()


def open_file(path: Path) -> None:
    """
    Open any given file with the default application.
    """
    logger.debug(f"Opening file {path}")
    try:
        # Use Qt to open the file, so that it works on all platforms.
        Qg.QDesktopServices.openUrl(Qc.QUrl.fromLocalFile(str(path)))
    except Exception:
        show_exception(None, tr("File Error"), tr("Failed to open file."))


def get_available_themes() -> list[tuple[str, str]]:
    """
    Check the data/color_themes directory for available themes.
    The theme name is the plain file name. The display name is either defined in the
    theme file under section General, key name.
    If not defined, the display name is the theme name but capitalized and
    with spaces instead of underscores.

    Note: The implicit system theme is not included in the list.

    :return: A list of available theme names with their display names.
    """
    # Simply discover all files in the themes folder.
    themes = []
    with resources.path(color_themes, "") as theme_dir:
        theme_dir = Path(theme_dir)
    for theme_file in theme_dir.iterdir():
        # Skip dirs and empty files.
        if theme_file.is_dir() or theme_file.stat().st_size == 0:
            continue
        theme_name = theme_file.stem
        content = theme_file.read_text()
        display_name = theme_name.replace("_", " ").capitalize()
        in_general_section = False
        for line in content.split("\n"):
            line = line.strip()
            if line.startswith("[General]"):
                in_general_section = True
            elif line.startswith("[") and line.endswith("]"):
                if in_general_section:
                    # We found general, but came across the next section now.
                    break
                in_general_section = False
            elif "=" in line:
                key, value = map(str.strip, line.split("=", 1))
                if key == "Name":
                    display_name = value
                    break
        themes.append((theme_name, display_name))

    return themes


def custom_icon_path(icon_name: str, theme: Literal["dark", "light"] | str = "") -> Path:
    """
    Loads the given icon from the dark, light, or color-agnostic set of custom icons.
    File names may omit the extension, in which case .svg and .png are checked.

    :param icon_name: The icon's filename, with or without extension.
    :param theme: Indicate if the icon should be pulled from the light or dark theme,
        if applicable, otherwise leave blank.
    :return: A full path to the file.
    """
    with resources.files(custom_icons) as data_path:
        custom_icon_dir: Path = data_path

    if theme:
        custom_icon_dir = custom_icon_dir / theme

    for extension in ("", ".svg", ".png"):
        icon_path = custom_icon_dir / (icon_name + extension)
        if icon_path.is_file():
            return icon_path

    raise FileNotFoundError(f"Failed to load '{custom_icon_dir / icon_name}'")


def load_custom_icon(icon_name: str, theme: Literal["dark", "light"] | str = "") -> Qg.QIcon:
    """
    Loads the given icon from the dark, light, or color-agnostic set of custom icons.
    File names may omit the extension, in which case .svg and .png are checked.
    If the file could not be found, a QIcon with a null pixmap is returned.

    :param icon_name: The icon's filename, with or without extension.
    :param theme: Indicate if the icon should be pulled from the light or dark theme,
        if applicable, otherwise leave blank.
    :return: A QIcon that may have a null pixmap.
    """
    try:
        icon_path = custom_icon_path(icon_name, theme)
        return Qg.QIcon(str(icon_path))
    except FileNotFoundError as e:
        logger.error(e)
        return Qg.QIcon()


class CaptureOutput(Qc.QObject):
    text_written = Qc.Signal(str, str)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._io_stdout = io.StringIO()
        self._io_stderr = io.StringIO()

    def __enter__(self) -> "CaptureOutput":
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr

    def write(self, string, stream="stdout") -> None:
        if stream == "stdout":
            self._io_stdout.write(string)
        elif stream == "stderr":
            self._io_stderr.write(string)
        self.text_written.emit(string, stream)

    def getvalue(self, stream="stdout") -> str:
        if stream == "stdout":
            return self._io_stdout.getvalue()
        elif stream == "stderr":
            return self._io_stderr.getvalue()

    def flush(self) -> None:
        self._io_stdout.flush()
        self._io_stderr.flush()


# Mapping between KDE color keys and QPalette roles
section_role_mapping = {
    "Colors:Button": {
        "BackgroundNormal": Qg.QPalette.Button,
        "ForegroundNormal": Qg.QPalette.ButtonText,
        "ForegroundActive": Qg.QPalette.BrightText,
    },
    "Colors:View": {
        "BackgroundNormal": Qg.QPalette.Base,
        "BackgroundAlternate": Qg.QPalette.AlternateBase,
        "ForegroundNormal": Qg.QPalette.Text,
        "ForegroundLink": Qg.QPalette.Link,
        "ForegroundVisited": Qg.QPalette.LinkVisited,
        "ForegroundInactive": Qg.QPalette.PlaceholderText,
    },
    "Colors:Selection": {
        "BackgroundNormal": Qg.QPalette.Highlight,
        "ForegroundNormal": Qg.QPalette.HighlightedText,
    },
    "Colors:Tooltip": {
        "BackgroundNormal": Qg.QPalette.ToolTipBase,
        "ForegroundNormal": Qg.QPalette.ToolTipText,
    },
    "Colors:Window": {
        "BackgroundNormal": Qg.QPalette.Window,
        "ForegroundNormal": Qg.QPalette.WindowText,
    },
    # Add more mappings as needed
}


def clamp_8bit(value: int) -> int:
    """
    Clamp a value to the 0-255 range.
    """
    return max(0, min(value, 255))


def apply_color_effect(
    source: Qg.QColor, effect_base: Qg.QColor, contrast_amount: float
) -> Qg.QColor:
    """
    Apply a color effect to a source color.

    :param source: The source color.
    :param effect_base: The base color for the effect.
    :param contrast_amount: The amount of contrast to apply.
    :return: The modified color.
    """
    # Essentially alpha blend, with the effect having the contrast amount as the alpha pasted on top.
    r = clamp_8bit(int(source.red() * (1 - contrast_amount) + effect_base.red() * contrast_amount))
    g = clamp_8bit(
        int(source.green() * (1 - contrast_amount) + effect_base.green() * contrast_amount)
    )
    b = clamp_8bit(
        int(source.blue() * (1 - contrast_amount) + effect_base.blue() * contrast_amount)
    )
    return Qg.QColor(r, g, b)


def load_color_palette(theme: str) -> Qg.QPalette:
    """
    Provide a theme name and get a QPalette object.
    The name should match one of the files in the themes folder.

    :param theme: The name of the theme.
    :return: A QPalette object.
    """
    palette = Qg.QPalette()

    with resources.files(color_themes) as data_path:
        file_path = data_path / theme

    file = Qc.QFile(file_path)
    if file.open(Qc.QFile.ReadOnly | Qc.QFile.Text):
        stream = Qc.QTextStream(file)
        content = stream.readAll()

        # Find the disabled color parameters.
        disabled_color = Qg.QColor.fromRgb(128, 128, 128)  # Default to gray.
        disabled_contrast_amount = 0.0
        in_disabled_section = False
        for line in content.split("\n"):
            line = line.strip()
            if line == "[ColorEffects:Disabled]":
                in_disabled_section = True
                continue
            if not in_disabled_section:
                continue
            if line.startswith("["):
                in_disabled_section = False
                break
            if "=" not in line:
                continue

            # Ok, now we're in the disabled section.
            key, value = map(str.strip, line.split("=", 1))
            if key == "Color":
                r, g, b = map(int, value.split(","))
                disabled_color.setRgb(r, g, b)
            elif key == "ContrastAmount":
                disabled_contrast_amount = float(value)

        section = None
        for line in content.split("\n"):
            line = line.strip()
            if line.startswith("[") and line.endswith("]"):
                section = line[1:-1]
            elif "=" in line:
                key, value = map(str.strip, line.split("=", 1))
                role_mapping = section_role_mapping.get(section, {})
                qt_color_role = role_mapping.get(key, None)
                if qt_color_role is not None:
                    r, g, b = map(int, value.split(","))
                    palette.setColor(Qg.QPalette.Normal, qt_color_role, Qg.QColor(r, g, b))
                    palette.setColor(Qg.QPalette.Inactive, qt_color_role, Qg.QColor(r, g, b))
                    # Calculate the disabled color.
                    disabled_color = apply_color_effect(
                        Qg.QColor(r, g, b), disabled_color, disabled_contrast_amount
                    )
                    palette.setColor(Qg.QPalette.Disabled, qt_color_role, disabled_color)
    else:
        raise ValueError(f"Could not open theme file: {theme}")

    # Fallback calculations
    if not palette.color(Qg.QPalette.Light).isValid():
        base = palette.color(Qg.QPalette.Button)
        palette.setColor(Qg.QPalette.Light, base.lighter(150))

    if not palette.color(Qg.QPalette.Dark).isValid():
        base = palette.color(Qg.QPalette.Window)
        palette.setColor(Qg.QPalette.Dark, base.darker(150))

    if not palette.color(Qg.QPalette.Midlight).isValid():
        light = palette.color(Qg.QPalette.Light)
        button = palette.color(Qg.QPalette.Button)
        midlight = Qg.QColor(
            (light.red() + button.red()) // 2,
            (light.green() + button.green()) // 2,
            (light.blue() + button.blue()) // 2,
        )
        palette.setColor(Qg.QPalette.Midlight, midlight)

    if not palette.color(Qg.QPalette.Mid).isValid():
        dark = palette.color(Qg.QPalette.Dark)
        button = palette.color(Qg.QPalette.Button)
        mid = Qg.QColor(
            (dark.red() + button.red()) // 2,
            (dark.green() + button.green()) // 2,
            (dark.blue() + button.blue()) // 2,
        )
        palette.setColor(Qg.QPalette.Mid, mid)

    if not palette.color(Qg.QPalette.Shadow).isValid():
        palette.setColor(Qg.QPalette.Shadow, Qg.QColor(0, 0, 0))

    return palette


def ansi_to_html(input_text: str) -> str:
    """
    Convert ansi escape sequences to html.

    Analytics HTML template
    <html><head><meta name="qrichtext" content="1" /><style type="text/css">
    p, li { white-space: pre-wrap; font-family: Noto Mono, Monospace; size: 12pt}
    </style></head><body>
    <p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px;
     -qt-block-indent:0; text-indent:0px;"><span style=" ">Mask Fitment Analytics
     <br />---------------------- <br />Total boxes: 5 | Masks succeeded: 5 (100%)
     | Masks failed: </span><span style="color:#b21818;">0</span><span style=" ">
     <br />Perfect masks: </span><span style="color:#18b2b2;">5</span><span style=" ">
     (100%) | Average border deviation: 0.00 <br /><br />Mask usage by mask size
     (smallest to largest): <br />Mask 0  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 1  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 2  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 3  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 4  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 5  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 6  : </span><span style="color:#18b2b2;">
     ████████████████████████████████████████</span><span style=" "> </span>
     <span style="color:#18b2b2;">1</span><span style=" "> / 1 <br />Mask 7  :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Mask 8  :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Mask 9  :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Mask 10 :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Box mask: </span>
     <span style="color:#18b2b2;">████████████████████████████████████████</span>
     <span style=" "> </span><span style="color:#18b2b2;">4</span><span style=" ">
     / 4 <br /><br /></span><span style="color:#18b2b2;">█ Perfect</span><span style=" ">
     | █ Total<br /><br /></span></p></body></html>
    """
    # Define ANSI escape sequences and corresponding Qt HTML color codes
    ansi_to_html_lut = {
        "31": "color:#eb514f;",
        "32": "color:#00ff00;",
        "92": "color:#22ff22;",
        "33": "color:#ffff00;",
        "34": "color:#0000ff;",
        "35": "color:#a771bf;",
        "36": "color:#00a3a3;",
        "37": "color:#ffffff;",
        "39": " ",  # Default color
    }

    # Initialize HTML header and footer
    html_header = """<html><head><meta name="qrichtext" content="1" /><style type="text/css">
        p, li { white-space: pre-wrap; font-family: Noto Mono, Monospace; size: 12pt}
        </style></head><body>
        <p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;">"""
    html_footer = "</p></body></html>"

    # Initialize output HTML with header
    buffer = StringIO()
    buffer.write(html_header)

    # Initialize current_style
    current_style = ""

    # Strip leading and trailing whitespace.
    input_text = input_text.strip()

    # Use regex to find all ANSI sequences and split text
    for plain_text, ansi_code in re.findall(r"(.*?)(?:\x1b\[(\d+)m|$)", input_text, re.DOTALL):
        # Add plain text
        newline = "\n"
        buffer.write(
            f'<span style="{current_style}">{plain_text.replace(newline, "<br />")}</span>'
        )

        # Update the style for the next span based on ANSI code
        if ansi_code:
            current_style = ansi_to_html_lut.get(ansi_code, "")

    # Add HTML footer
    buffer.write(html_footer)

    return buffer.getvalue()


# Map the OCR parse error codes to i18n-able strings.
OCR_PARSE_ERROR_CODES = {
    op.ParseErrorCode.OS_ERROR: tr("Failed to access the file."),
    op.ParseErrorCode.OTHER_ERROR: tr("An error occurred."),
    op.ParseErrorCode.NOT_6_COLUMNS: tr("The CSV file format requires exactly 6 columns."),
    op.ParseErrorCode.NOT_AN_INT: tr("The coordinates must be integers."),
    op.ParseErrorCode.INT_TOO_BIG: tr("The box coordinates may not exceed 2 billion."),
    op.ParseErrorCode.INVALID_FORMAT: tr("The file format was not recognized."),
    op.ParseErrorCode.INVALID_CSV_HEADER: tr(
        "The CSV file must start with a header row, followed by data rows."
    ),
    op.ParseErrorCode.NO_FILE_PATH: tr("The file path is missing."),
    op.ParseErrorCode.INVALID_PATH: tr("The file path is invalid."),
}


def format_ocr_parse_errors(errors: list[op.ParseError]) -> str:
    """
    Format the OCR parse errors into a human-readable, i18n-able string.

    ParseErrors contain the following:
    - line: int
    - error_code: ParseErrorCode
    - context: str

    :param errors: The list of OCR parse errors.
    :return: A formatted string.
    """

    buffer = StringIO()
    buffer.write(tr("OCR Parse Errors"))
    buffer.write("\n\n")

    for error in errors:
        buffer.write(f"- Line {error.line}: ")
        buffer.write(OCR_PARSE_ERROR_CODES[error.error_code])
        buffer.write("\n")
        buffer.write(f"  {error.context}\n\n")

    return buffer.getvalue()


class AmbiguousPath(Exception):
    pass


def match_image_files_to_ocr_analytics(
    images: list[imf.ImageFile], analytics: list[st.OCRAnalytic]
) -> tuple[dict[imf.ImageFile, st.OCRAnalytic], list[imf.ImageFile], list[st.OCRAnalytic]]:
    """
    Attempt to match all the currently loaded images to the OCR analytics.
    If an image couldn't find a partner, it gets added to the unmatched list of Images.
    Likewise, unmatched analytics are added to the unmatched list of Paths.

    If an analytic could be assigned to multiple images, raise an AmbiguousPath exception.

    :param images: The list of ImageFiles.
    :param analytics: The list of OCRAnalytics.
    :return: A dictionary of matched ImageFiles and OCRAnalytics, and two lists of unmatched items.
    """
    matched = {}
    unmatched_images = images.copy()
    unmatched_analytics = analytics.copy()

    # Attempt to match the file names in the analytics to the image paths
    # by making sure the file path is a suffix of the image path.
    for analytic in analytics:
        # Assumption: all removed box data elements report the same path,
        # so choose the first one.
        file_path_fragment = analytic.path

        for image in images:
            if str(image.path).endswith(str(file_path_fragment)):
                # Guard against double assignments.
                if image in matched:
                    raise AmbiguousPath(
                        tr("Multiple analytics match the image: {path}").format(path=image.path)
                    )
                if analytic not in unmatched_analytics:
                    raise AmbiguousPath(
                        tr("Multiple images match the file path: {path}").format(
                            path=file_path_fragment
                        )
                    )
                matched[image] = analytic
                unmatched_images.remove(image)
                unmatched_analytics.remove(analytic)

    return matched, unmatched_images, unmatched_analytics



import re
from importlib import resources

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw

from pcleaner import data

colors_dark_theme = {
    "debug": Qg.QColor(41, 128, 185),
    "info": Qg.QColor(39, 174, 96),
    "warning": Qg.QColor(196, 91, 0),
    "error": Qg.QColor(218, 68, 83),
    "critical": Qg.QColor(149, 218, 76),
    "critical_bg": Qg.QColor(77, 31, 36),
    "traceback_values": Qg.QColor(40, 160, 160),
}
colors_light_theme = {
    "debug": Qg.QColor(0, 87, 174),
    "info": Qg.QColor(0, 110, 40),
    "warning": Qg.QColor(176, 128, 0),
    "error": Qg.QColor(191, 3, 3),
    "critical": Qg.QColor(191, 82, 3),
    "critical_bg": Qg.QColor(247, 230, 230),
    "traceback_values": Qg.QColor(0, 128, 128),
}


class Highlighter(Qg.QSyntaxHighlighter):
    def __init__(self, parent=None):
        Qg.QSyntaxHighlighter.__init__(self, parent)

        self._mappings = {}

    def add_mapping(self, pattern, format_):
        self._mappings[pattern] = format_

    def highlightBlock(self, text):
        for pattern, format_ in self._mappings.items():
            for match in re.finditer(pattern, text):
                start, end = match.span()
                self.setFormat(start, end - start, format_)


class LogSyntaxManager(Qc.QObject):
    def __init__(self, parent, widget: Qw.QPlainTextEdit, palette: Qg.QPalette):
        Qc.QObject.__init__(self, parent)

        self._highlighter = Highlighter()

        self.setup_editor(widget, palette)

    def setup_editor(self, widget: Qw.QPlainTextEdit, palette: Qg.QPalette):
        background_color = palette.color(Qg.QPalette.Window)
        theme_is_dark = background_color.lightness() < 128

        if theme_is_dark:
            colors = colors_dark_theme
        else:
            colors = colors_light_theme

        debug_format = Qg.QTextCharFormat()
        debug_format.setForeground(Qg.QBrush(colors["debug"]))

        info_format = Qg.QTextCharFormat()
        info_format.setForeground(Qg.QBrush(colors["info"]))

        warning_format = Qg.QTextCharFormat()
        warning_format.setForeground(Qg.QBrush(colors["warning"]))

        error_format = Qg.QTextCharFormat()
        error_format.setForeground(Qg.QBrush(colors["error"]))
        error_format.setFontWeight(Qg.QFont.Bold)

        critical_format = Qg.QTextCharFormat()
        critical_format.setForeground(Qg.QBrush(colors["critical"]))
        critical_format.setBackground(Qg.QBrush(colors["critical_bg"]))
        critical_format.setFontWeight(Qg.QFont.Bold)

        traceback_values_format = Qg.QTextCharFormat()
        traceback_values_format.setForeground(Qg.QBrush(colors["traceback_values"]))

        traceback_file_format = Qg.QTextCharFormat()
        traceback_file_format.setForeground(Qg.QBrush(colors["error"]))

        template = r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3} \| LEVEL +\| [\w\.]+:[\w]+:\d+ -"

        self._highlighter.add_mapping(template.replace("LEVEL", "DEBUG"), debug_format)
        self._highlighter.add_mapping(template.replace("LEVEL", "INFO"), info_format)
        self._highlighter.add_mapping(template.replace("LEVEL", "WARNING"), warning_format)
        self._highlighter.add_mapping(template.replace("LEVEL", "ERROR"), error_format)
        self._highlighter.add_mapping(template.replace("LEVEL", "CRITICAL"), critical_format)
        self._highlighter.add_mapping(r"^\W*[└│].*$", traceback_values_format)
        self._highlighter.add_mapping(r'^>?\s*File ".*?", line \d+, in .+$', traceback_file_format)

        self._highlighter.setDocument(widget.document())


class LogViewer(Qw.QPlainTextEdit):
    """
    Show logs of individual sessions.
    Visualize them with syntax highlighting and stuff.
    """

    syntax_highlighter: LogSyntaxManager

    def __init__(
        self,
        parent=None,
    ) -> None:
        """
        Init the widget.

        :param parent: The parent widget.
        """
        Qw.QPlainTextEdit.__init__(self, parent)

        with resources.files(data) as data_path:
            font_path = data_path / "NotoMono-Regular.ttf"
        font_id = Qg.QFontDatabase.addApplicationFont(str(font_path))
        self.setFont(Qg.QFont(Qg.QFontDatabase.applicationFontFamilies(font_id)[0]))
        self.syntax_highlighter = LogSyntaxManager(self, self, self.palette())

    def show_log(self, log_text: str) -> None:
        """
        Show the log in the text edit.

        :param log_text: The log text.
        """
        self.setPlainText(log_text)

    def get_log(self) -> str:
        """
        Get the log text.

        :return: The log text.
        """
        return self.toPlainText()

    def scroll_to_bottom(self) -> None:
        """
        Scroll to the bottom of the log.
        """
        self.verticalScrollBar().setValue(self.verticalScrollBar().maximum())



import locale as pylocale
import os
import platform
import sys
from importlib import resources
from io import StringIO

import PySide6
import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
import torch
from PIL import Image
from docopt import docopt
from loguru import logger

import pcleaner.cli_utils as cu
import pcleaner.config as cfg
import pcleaner.data.theme_icons as theme_icons_data
import pcleaner.data.translation_generated_files as translation_data
import pcleaner.gui.gui_utils as gu
import pcleaner.helpers as hp
from pcleaner import __display_name__, __version__
from pcleaner.gui.mainwindow_driver import MainWindow


# Allow loading of large images.
Image.MAX_IMAGE_PIXELS = 2**32


def launch(files_to_open: list[str], debug: bool = False) -> None:
    """
    Launch the GUI.

    :param files_to_open: A list of files to open.
    :param debug: Whether to enable debug mode.
    """

    cu.get_log_path().parent.mkdir(parents=True, exist_ok=True)

    # Set up file logging.
    logger.add(str(cu.get_log_path()), rotation="10 MB", retention="1 week", level="DEBUG")

    # Set up a preliminary exception handler so that this still shows up in the log.
    # Once the gui is up and running it'll be replaced with a call to the gui's error dialog.
    def exception_handler(exctype, value, traceback) -> None:
        logger.opt(depth=1, exception=(exctype, value, traceback)).critical(
            "An uncaught exception was raised before the GUI was initialized."
        )

    sys.excepthook = exception_handler

    logger.info("\n" + cfg.STARTUP_MESSAGE)
    buffer = StringIO()
    buffer.write("\n- Program Information -\n")
    buffer.write(f"Program: {__display_name__} {__version__}\n")
    buffer.write(f"Executing from: {__file__}\n")
    buffer.write(f"Log file: {cu.get_log_path()}\n")
    buffer.write(f"Config file: {cu.get_config_path()}\n")
    buffer.write(f"Cache directory: {cu.get_cache_path()}\n")
    buffer.write("- System Information -\n")
    buffer.write(f"Operating System: {platform.system()} {platform.release()}\n")
    if platform.system() == "Linux":
        buffer.write(f"Desktop Environment: {os.getenv('XDG_CURRENT_DESKTOP', 'unknown')}\n")
    buffer.write(f"Python Version: {sys.version}\n")
    buffer.write(f"PySide (Qt) Version: {PySide6.__version__}\n")
    buffer.write(f"Available Qt Themes: {', '.join(Qw.QStyleFactory.keys())}\n")
    current_app_theme = Qw.QApplication.style()
    current_app_theme_name = (
        current_app_theme.objectName() if current_app_theme else "System Default"
    )
    buffer.write(f"Current Qt Theme: {current_app_theme_name}\n")
    icon_theme_name = Qg.QIcon.themeName()
    icon_theme_name = icon_theme_name if icon_theme_name else "System Default"
    buffer.write(f"Current Icon Theme: {icon_theme_name}\n")
    buffer.write(f"System locale: {Qc.QLocale.system().name()}\n")
    buffer.write(f"Architecture: {platform.machine()}\n")
    buffer.write(f"CPU Cores: {os.cpu_count()}\n")
    buffer.write(f"Memory: {hp.sys_virtual_memory_total() / 1024 ** 3:.2f} GiB\n")
    buffer.write(f"Swap: {hp.sys_swap_memory_total() / 1024 ** 3:.2f} GiB\n")
    if torch.cuda.is_available():
        buffer.write(f"GPU: {torch.cuda.get_device_name(0)} (CUDA enabled)\n")
        buffer.write(f"    CUDA Version: {torch.version.cuda}\n")
        buffer.write(
            f"    CUDA Cores: {torch.cuda.get_device_properties(0).multi_processor_count}\n"
        )
        buffer.write(
            f"    VRAM: {torch.cuda.get_device_properties(0).total_memory / 1024 ** 3:.2f} GiB\n"
        )
    else:
        buffer.write("GPU: None (CUDA not available)\n")

    logger.info(buffer.getvalue())

    # Start the main window.
    app = Qw.QApplication(sys.argv)

    # Load the config.
    config = cfg.load_config()

    # Apply the locale from the config.
    if config.locale:
        locale = Qc.QLocale(config.locale)
        # Apply it to python as well.
        pylocale.setlocale(pylocale.LC_ALL, locale.name())

    else:
        locale = Qc.QLocale.system()

    logger.info(f"Using locale {locale.name()}.")

    translator = Qc.QTranslator(app)

    # Load translations.
    path = Qc.QLibraryInfo.location(Qc.QLibraryInfo.TranslationsPath)
    if translator.load(locale, "qt", "_", path):
        logger.debug(f"Loaded built-in Qt translations for {locale.name()}.")
        app.installTranslator(translator)
    else:
        logger.warning(f"Failed to load Qt translations from {path}.")

    translator = Qc.QTranslator(app)

    if translator.load(locale, "qtbase", "_", path):
        logger.debug(f"Loaded built-in Qt base translations for {locale.name()}.")
        app.installTranslator(translator)
    else:
        logger.warning(f"Failed to load Qt base translations from {path}.")

    translator = Qc.QTranslator(app)

    with resources.files(translation_data) as data_path:
        path = str(data_path)

    if translator.load(locale, "PanelCleaner", "_", path):
        app.installTranslator(translator)
        logger.debug(f"Loaded App translations for {locale.name()}.")

    with resources.files(theme_icons_data) as data_path:
        theme_icons = str(data_path)

    Qg.QIcon.setFallbackSearchPaths([":/icons", theme_icons])
    # We need to set an initial theme on Windows, otherwise the icons will fail to load
    # later on, even when switching the theme again.
    if platform.system() == "Windows":
        Qg.QIcon.setThemeName("breeze")
        Qg.QIcon.setThemeSearchPaths([":/icons", theme_icons])

    if files_to_open:
        input_paths = ", ".join(map(str, files_to_open))
        logger.info(f"Files to open: {input_paths}")

    try:
        window = MainWindow(config, files_to_open, debug)
        window.show()
        sys.exit(app.exec())
    except Exception:
        gu.show_exception(None, "Failed to launch", "Failed to initialize the main window.")
    finally:
        logger.info(cfg.SHUTDOWN_MESSAGE + "\n")


def main():
    docopt_doc = """Panel Cleaner

    Usage:
        pcleaner-gui [<image_path> ...] [--debug]

    Options:
        <image_path>          One or multiple files or directories to load on startup.
        --debug               Enable debug mode.
    """
    args = docopt(docopt_doc, version=f"Panel Cleaner {__version__}")
    launch(args.image_path, args.debug)


if __name__ == "__main__":
    main()
